###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""UploadMC module is used to upload to ES the json files for MC statistics."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import io
import json

import six
from DIRAC import S_OK, S_ERROR, gLogger
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.ProductionManagementSystem.Client.MCStatsClient import MCStatsClient
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary
from LHCbDIRAC.Core.Utilities.GeneratorLog import GeneratorLog


class UploadMC(ModuleBase):
  """Upload to LogSE."""

  def __init__(self):
    """Module initialization."""

    self.log = gLogger.getSubLogger("UploadMC")
    super(UploadMC, self).__init__(self.log)

    self.version = __RCSID__

  def _resolveInputVariables(self):
    """standard method for resolving the input variables."""

    super(UploadMC, self)._resolveInputVariables()

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_number=None, step_id=None):
    """ Main executon method.

        Takes care of indexing JSON files on ElasticSearch databases
        All these uploads are controlled by flags that can be found in
        'Operations/<setup>/Productions/'

        1) Gauss errors => /Operations/<setup>/Productions/UploadES_GaussErrors (default: True)
        2) Boole errors => /Operations/<setup>/Productions/UploadES_BooleErrors (default: True)
        3) Summary.xml -> JSON => /Operations/<setup>/Productions/UploadES_XMLSummary (default: False)
        4) GeneratorLog.xml -> JSON => /Operations/<setup>/Productions/UploadES_GeneratorLog (default: False)
        5) PRMon => /Operations/<setup>/Productions/UploadES_PrMon (default: False)

    """
    try:

      super(UploadMC, self).execute(self.version, production_id,
                                    prod_job_id, wms_job_id,
                                    workflowStatus, stepStatus,
                                    wf_commons, step_commons,
                                    step_number, step_id)

      self._resolveInputVariables()

      mcStatsClient = MCStatsClient()

      # DBs list
      # db = {
      #     'XMLSummary': elasticApplicationSummaryDB,
      #     'booleErrors': elasticMCBooleLogErrorsDB,
      #     'gaussErrors': elasticMCGaussLogErrorsDB,
      #     'generatorLog': elasticMCGeneratorLogDB,
      #     'prMon': elasticMCGeneratorLogDB,
      # }

      # 1) MC Gauss errors
      # looking for json files that are 'self.jobID_Errors_Gauss.json'
      fn = '%s_Errors_Gauss.json' % (self.jobID)
      if os.path.exists(fn):
        with io.open(fn) as fd:
          try:
            jsonData = json.load(fd)
            self.log.verbose("Content of JSON file", "%s: %s" % (fn, jsonData))
            if self._enableModule() and self.opsH.getValue('Productions/UploadES_GaussErrors', True):
              res = mcStatsClient.set('gaussErrors', jsonData)
              if not res['OK']:
                self.log.error('MC Error data not set, exiting without affecting workflow status',
                               "%s: %s" % (str(jsonData), res['Message']))
            else:
              # At this point we can see exactly what the module would have uploaded
              self.log.info("Module disabled", "would have attempted to upload the following file %s" % fn)
          except Exception as ve:
            self.log.error(repr(ve))
            self.log.verbose("Exception loading the JSON file: content of %s follows" % fn)
            self.log.verbose(fd.read)
            # do not fail the job for this
            # raise
      else:
        self.log.info("Gauss errors JSON file not found", fn)

      # 2) MC Boole errors
      # looking for json files that are 'self.jobID_Errors_Boole.json'
      fn = '%s_Errors_Boole.json' % (self.jobID)
      if os.path.exists(fn):
        with io.open(fn) as fd:
          try:
            jsonData = json.load(fd)
            self.log.verbose("Content of JSON file", "%s: %s" % (fn, jsonData))
            if self._enableModule() and self.opsH.getValue('Productions/UploadES_BooleErrors', True):
              res = mcStatsClient.set('booleErrors', jsonData)
              if not res['OK']:
                self.log.error('MC Error data not set, exiting without affecting workflow status',
                               "%s: %s" % (str(jsonData), res['Message']))
            else:
              # At this point we can see exactly what the module would have uploaded
              self.log.info("Module disabled", "would have attempted to upload the following file %s" % fn)
          except Exception as ve:
            self.log.error(repr(ve))
            self.log.verbose("Exception loading the JSON file: content of %s follows" % fn)
            self.log.verbose(fd.read)
            # do not fail the job for this
            # raise
      else:
        self.log.info("Boole errors JSON file not found", fn)

      # 3) summary JSON files
      # looking for xml files that are 'summaryGauss_self.production_id_self.prod_job_id_1.xml'
      xmlfl = 'summaryGauss_%s_%s_1.xml' % (self.production_id, self.prod_job_id)
      if os.path.exists(xmlfl):

        try:
          xmlData = XMLSummary(xmlfl)
          xmlData.xmltojson()
          # At this point 'summaryGauss_self.production_id_self.prod_job_id_1.json' should have been created
          jsonfl = 'summaryGauss_%s_%s_1.json' % (self.production_id, self.prod_job_id)
          with io.open(jsonfl) as JS:
            jsonData = json.load(JS)
            ids = dict()
            ids['JobID'] = self.jobID
            ids['ProductionID'] = self.production_id
            ids['prod_job_id'] = self.prod_job_id
            jsonData['Counters']['ID'] = ids
            with io.open(jsonfl, 'w', encoding="utf-8") as output:
              output.write(six.text_type(json.dumps(jsonData, indent=2)))

            self.log.verbose("Content of JSON file", "%s: %s" % (jsonfl, jsonData))
            if self._enableModule() and self.opsH.getValue('Productions/UploadES_XMLSummary', False):
              res = mcStatsClient.set('XMLSummary', jsonData)
              if not res['OK']:
                self.log.error('Gauss Summaries data not set, exiting without affecting workflow status',
                               "%s: %s" % (str(jsonData), res['Message']))
            else:
              # At this point we can see exactly what the module would have uploaded
              self.log.info("Module disabled", "would have attempted to upload the following file %s" % jsonfl)

        except Exception:
          self.log.exception("Exception creating/loading the XMLSummary JSON file")
          # do not fail the job for this

      else:
        self.log.info("XML Gauss summary file not found", xmlfl)

      # 4) Generator logs
      # looking for xml files that are 'GeneratorLog.xml'
      xmlfile = 'GeneratorLog.xml'
      if os.path.exists(xmlfile):

        try:
          jsonfile = 'GeneratorLog_%s_%s.json' % (self.production_id, self.prod_job_id)
          xmlData = GeneratorLog()
          xmlData.generatorLogJson(jsonfile)
          # At this point 'GeneratorLog_self.production_id_self.prod_job_id.json' should have been created
          with io.open(jsonfile) as JS:
            jsonData = json.load(JS)
            ids = dict()
            ids['JobID'] = self.jobID
            ids['ProductionID'] = self.production_id
            ids['prod_job_id'] = self.prod_job_id
            jsonData['generatorCounters']['ID'] = ids
            with io.open(jsonfile, 'w', encoding="utf-8") as output:
              output.write(six.text_type(json.dumps(jsonData)))

            self.log.verbose("Content of JSON file", "%s: %s" % (jsonfile, jsonData))
            if self._enableModule() and self.opsH.getValue('Productions/UploadES_GeneratorLog', False):
              res = mcStatsClient.set('generatorLog', jsonData)
              if not res['OK']:
                self.log.error('Generator Log data not set, exiting without affecting workflow status',
                               "%s: %s" % (str(jsonData), res['Message']))
            else:
              # At this point we can see exactly what the module would have uploaded
              self.log.info("Module disabled", "would have attempted to upload the following file %s" % jsonfile)

        except Exception:
          self.log.exception("Exception creating/loading the GeneratorLog JSON file")
          # do not fail the job for this

      else:
        self.log.info("XML GeneratorLog file not found", xmlfile)

      # 5) PRmon JSON files (only for Gauss)
      # looking for prmon files that are 'prmon_Gauss.json'
      prmonFile = 'prmon_Gauss.json'
      if os.path.exists(prmonFile):
        with io.open(prmonFile) as JS:
          try:
            jsonData = json.load(JS)
            jsonData['JobID'] = self.jobID
            jsonData['ProductionID'] = self.production_id
            jsonData['prod_job_id'] = self.prod_job_id
            self.log.verbose("Content of JSON file", "%s: %s" % (prmonFile, jsonData))
            if self._enableModule() and self.opsH.getValue('Productions/UploadES_PrMon', False):
              res = mcStatsClient.set('prMon', jsonData)
              if not res['OK']:
                self.log.error('prmon Metrics data not set, exiting without affecting workflow status',
                               "%s: %s" % (str(jsonData), res['Message']))
            else:
              # At this point we can see exactly what the module would have uploaded
              self.log.info("Module disabled", "would have attempted to upload the following file %s" % prmonFile)
          except Exception as ve:
            self.log.error(repr(ve))
            self.log.verbose("Exception loading the JSON file: content of %s follows" % prmonFile)
            self.log.verbose(JS.read())
            # do not fail the job for this
            # raise
      else:
        self.log.info("prmon JSON file not found", prmonFile)

      return S_OK()

    except Exception as e:
      self.log.exception("Failure in UploadMC execute module", lException=e)
      return S_ERROR(repr(e))

    finally:
      super(UploadMC, self).finalize(self.version)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def usage():
  """usage.

  Prints script usage
  """
  from DIRAC.Core.Base import Script

  print('Usage: %s <LFN> [<LFN>] [--ProductionID=<ID>] [Try -h,--help for more information]' % Script.scriptName)
  DIRAC.exit(2)


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.registerSwitch(
      "",
      "ProductionID=",
      "Restrict query to given production ID (default is to show status for all)",
  )
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

  prodID = ''
  for switch in Script.getUnprocessedSwitches():
    if switch[0].lower() == "productionid":
      prodID = switch[1]

  args = Script.getPositionalArgs()

  if len(args) < 1:
    usage()

  if prodID:
    try:
      prodID = int(prodID)
    except Exception as x:
      print('ERROR ProductionID should be an integer')
      DIRAC.exit(2)

  diracProd = DiracProduction()
  exitCode = 0
  result = diracProd.checkFilesStatus(args, prodID, printOutput=False)
  if not result['OK']:
    print('ERROR %s' % (result['Message']))
    exitCode = 2

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Use dirac-proxy-init to get a proxy."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import sys
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  if os.getenv('X509_CERT_DIR') is None:
    sys.exit('the variable X509_CERT_DIR do not exist')

  if not os.path.isdir(os.environ['X509_CERT_DIR']):
    sys.exit('the directory %s does not exist' % os.environ['X509_CERT_DIR'])

  if os.getenv('X509_VOMS_DIR') is None:
    sys.exit('the variable X509_VOMS_DIR do not exist')

  if not os.path.isdir(os.environ['X509_VOMS_DIR']):
    sys.exit('the directory %s does not exist' % os.environ['X509_VOMS_DIR'])

  out = os.system("dirac-proxy-init -o LogLevel=NOTICE --strict --rfc '%s'" % "' '".join(sys.argv[1:]))
  sys.exit(int(out / 256))


if __name__ == "__main__":
  main()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the destination for a set of runs, based on the majority of reco
output."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def execute(dmScript):
  """Parse the options and execute the script."""
  from DIRAC import gLogger
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar

  bkQuery = dmScript.getBKQuery()
  fileType = bkQuery.getFileTypeList()
  if not set(fileType) & {'FULL.DST', 'RDST', 'SDST'}:
    gLogger.error("Please provide a reconstruction BK path")
    DIRAC.exit(1)

  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  from DIRAC.DataManagementSystem.Client.DataManager import DataManager
  from DIRAC.Core.Utilities.List import breakListIntoChunks
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup

  bk = BookkeepingClient()
  tr = TransformationClient()
  dm = DataManager()
  dmsHelper = DMSHelpers()

  bkQueryDict = bkQuery.getQueryDict()
  gLogger.notice("For BK Query:", str(bkQueryDict))
  progressBar = ProgressBar(1, title="Running BK query...", step=1)
  res = bk.getFilesWithMetadata(bkQueryDict)
  if not res['OK']:
    gLogger.error("Error getting files from BK", res['Message'])
    DIRAC.exit(2)

  if 'ParameterNames' in res.get('Value', {}):
    parameterNames = res['Value']['ParameterNames']
    info = res['Value']['Records']
    progressBar.endLoop("Obtained %d files" % len(info))
  else:
    gLogger.error('\nNo metadata found')
    DIRAC.exit(3)
  lfns = []
  runLFNs = {}
  for item in info:
    metadata = dict(zip(parameterNames, item))
    lfn = metadata['FileName']
    lfns.append(lfn)
    runLFNs.setdefault(metadata['RunNumber'], []).append(lfn)

  chunkSize = 1000
  progressBar = ProgressBar(len(lfns), title='Getting replicas of %d files' % len(lfns), chunk=chunkSize)
  replicas = {}
  errors = {}
  for lfnChunk in breakListIntoChunks(lfns, chunkSize):
    progressBar.loop()
    res = dm.getReplicas(lfnChunk, getUrl=False)
    if not res['OK']:
      errors.setdefault(res['Message'], []).extend(lfnChunk)
    else:
      replicas.update(res['Value']['Successful'])
      for lfn, error in res['Value']['Failed'].items():
        errors.setdefault(error, []).append(lfn)
  progressBar.endLoop()
  for error, lfns in errors.items():
    gLogger.error(error, 'for %d files' % len(lfns))

  tier1RDST = set(resolveSEGroup('Tier1-RDST'))
  setOK = 0
  errors = {}
  progressBar = ProgressBar(len(runLFNs), title='Defining destination for %d runs' % len(runLFNs), step=10)
  for run, lfns in runLFNs.items():
    progressBar.loop()
    res = tr.getDestinationForRun(run)
    if res.get('Value'):
      errors.setdefault('Destination already set', []).append(str(run))
      continue
    # print 'Run', run, len( lfns ), 'Files', lfns[:3]
    seCounts = {}
    for lfn in lfns:
      for se in tier1RDST.intersection(replicas.get(lfn, [])):
        seCounts[se] = seCounts.setdefault(se, 0) + 1
    # print seCounts
    maxi = 0
    seMax = None
    for se, count in seCounts.items():
      if count > maxi:
        seMax = se
        maxi = count
    if not seMax:
      errors.setdefault('No SE found, use CERN-RDST', []).append(str(run))
      seMax = 'CERN-RDST'
    # SE found, get its site
    res = dmsHelper.getLocalSiteForSE(seMax)
    if res['OK']:
      site = res['Value']
      res = tr.setDestinationForRun(run, site)
      if not res['OK']:
        errors.setdefault(res['Message'], []).append(str(run))
      else:
        setOK += 1
  progressBar.endLoop('Successfully set destination for %d runs' % setOK)
  for error, runs in errors.items():
    gLogger.error(error, 'for runs %s' % ','.join(runs))


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

  dmScript = DMScript()
  dmScript.registerBKSwitches()

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile]' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=False)

  execute(dmScript)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set files Removed in a transformation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def __getTransformations(args):
  from DIRAC.Core.Base import Script
  from DIRAC import gLogger

  transList = []
  if not len(args):
    print("Specify transformation number...")
    Script.showHelp()
  else:
    ids = args[0].split(",")
    try:
      for transID in ids:
        r = transID.split(':')
        if len(r) > 1:
          for i in range(int(r[0]), int(r[1]) + 1):
            transList.append(i)
        else:
          transList.append(int(r[0]))
    except Exception as e:
      gLogger.exception("Invalid transformation", lException=e)
      transList = []
  return transList


@DIRACScript()
def main():
  import DIRAC
  from DIRAC import gLogger
  from DIRAC.Core.Base import Script
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

  dmScript = DMScript()
  dmScript.registerFileSwitches()

  Script.parseCommandLine(ignoreErrors=True)

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  runInfo = True
  userGroup = None

  transList = __getTransformations(Script.getPositionalArgs())
  if not transList:
    DIRAC.exit(1)

  requestedLFNs = dmScript.getOption('LFNs')
  if not requestedLFNs:
    gLogger.always('No files to add')
    DIRAC.exit(1)

  from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  trClient = TransformationClient()
  rc = 0
  for transID in transList:
    res = trClient.setFileStatusForTransformation(transID, 'Removed', requestedLFNs, force=True)
    if res['OK']:
      gLogger.always(
          'Successfully set %d files%s Removed in transformation %d' %
          (len(
              res['Value']), (' (out of %d)' %
                              len(requestedLFNs)) if len(
              res['Value']) != len(requestedLFNs) else '', transID))
    else:
      gLogger.always('Failed to set %d files Removed in transformation %d' %
                     (len(requestedLFNs), transID), res['Message'])
      rc = 2
  DIRAC.exit(rc)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The Storage History Agent will create a summary of the storage usage DB
grouped by processing pass or other interesting parameters.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time
import copy
import datetime
from collections import defaultdict
import six

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.File import mkDir, convertSizeUnits
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities.List import breakListIntoChunks

from LHCbDIRAC.AccountingSystem.Client.Types.UserStorage import UserStorage
from LHCbDIRAC.AccountingSystem.Client.Types.Storage import Storage
from LHCbDIRAC.AccountingSystem.Client.Types.DataStorage import DataStorage
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.DataUsageClient import DataUsageClient
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

__RCSID__ = "$Id$"


def _standardDirectory(dirPath):
  """ Add a "/" at the end of the directory name if not present """
  return os.path.join(dirPath, '')


def _fillMetadata(dictToFill, metadataValue):
  """Fill the dictionary to send to the accounting.

  If metadataValue is a string then set all the values of dictToFill, to
  this value if metadataValue is a dictionary then set each value of
  dictToFill to the corresponding value of metadataValue
  """
  # ds = DataStorage()
  # keyList = ds.keyFieldsList
  # this is the list of attributes returned by the Bookkeeping for a given directory
  keyList = ('ConfigName', 'ConfigVersion', 'FileType', 'Production',
             'ProcessingPass', 'ConditionDescription', 'EventType', 'Visibility')
  if isinstance(metadataValue, six.string_types):
    for k in keyList:
      dictToFill[k] = metadataValue
  elif isinstance(metadataValue, dict):
    for k in keyList:
      dictToFill[k] = metadataValue.get(k, 'na')


class StorageHistoryAgent(AgentModule):
  def initialize(self):
    """Sets defaults."""
    self.am_setOption('PollingTime', 43200)
    if self.am_getOption('DirectDB', False):
      from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB
      self.__stDB = StorageUsageDB()
    else:
      self.__stDB = StorageUsageClient()
    self.__workDirectory = self.am_getOption("WorkDirectory")
    mkDir(self.__workDirectory)
    self.log.info("Working directory is %s" % self.__workDirectory)

    self.__ignoreDirsList = self.am_getOption('Ignore', [])
    self.log.info("List of directories to ignore for the DataStorage accounting: %s " % self.__ignoreDirsList)

    self.__bkClient = BookkeepingClient()
    self.__dataUsageClient = DataUsageClient()
    self.__fileCatalog = FileCatalog()
    self.cachedMetadata = {}
    # build a dictionary with Event Type descriptions (to be send to accounting, instead of number Event Type ID)
    self.eventTypeDescription = {'na': 'na',
                                 'notInBkk': 'notInBkk',
                                 'FailedBkkQuery': 'FailedBkkQuery',
                                 'None': 'None'}
    self.limitForCommit = self.am_getOption("LimitForCommit", 1000)
    self.callsToGetSummary = 0
    self.callsToDirectorySummary = 0
    self.callsToDudbForMetadata = 0
    # keep count of calls to Bkk
    self.callsToBkkgetDirectoryMetadata = 0
    self.callsToBkkGetEvtType = 0

    return S_OK()

  def userStorageAccounting(self):
    self.log.notice("-------------------------------------------------------------------------------------\n")
    self.log.notice("Generate accounting records for user directories ")
    self.log.notice("-------------------------------------------------------------------------------------\n")

    result = self.__stDB.getUserSummary()
    if not result['OK']:
      return result
    userCatalogData = result['Value']
    self.log.notice("Got summary for %s users" % (len(userCatalogData)))
    result = self.__stDB.getUserSummaryPerSE()
    if not result['OK']:
      return result
    userSEData = result['Value']
    self.log.notice("Got SE summary for %s users" % (len(userSEData)))

    now = datetime.datetime.utcnow()
    numRows = 0
    for user in sorted(userSEData):
      if user not in userCatalogData:
        self.log.error("User has SE data but not Catalog data!", user)
        continue
      for se in sorted(userSEData[user]):
        seData = userSEData[user][se]
        usRecord = UserStorage()
        usRecord.setStartTime(now)
        usRecord.setEndTime(now)
        usRecord.setValueByKey("User", user)
        usRecord.setValueByKey("StorageElement", se)
        usRecord.setValueByKey("LogicalSize", userCatalogData[user]['Size'])
        usRecord.setValueByKey("LogicalFiles", userCatalogData[user]['Files'])
        usRecord.setValueByKey("PhysicalSize", seData['Size'])
        usRecord.setValueByKey("PhysicalFiles", seData['Files'])
        usRecord.setValueByKey("StorageSize", 0)
        usRecord.setValueByKey("StorageFiles", 0)
        gDataStoreClient.addRegister(usRecord)
        numRows += 1

      self.log.notice(" User %s is using %.2f GiB (%s files)" % (user,
                                                                 userCatalogData[user]['Size'] / (1024.0 ** 3),
                                                                 userCatalogData[user]['Files']))
    self.log.notice("Sending %s records to accounting for user storage" % numRows)
    res = gDataStoreClient.commit()
    if not res['OK']:
      self.log.notice("ERROR: committing UserStorage records: %s " % res)
      return S_ERROR(res)
    else:
      self.log.notice("%s records for UserStorage type successfully committed" % numRows)

  def __getTopDirUsage(self, dirToScan, topDirLogicalUsage):
    """
    Get logical storage information from su_Directory table
    """
    res = self.__fileCatalog.listDirectory(dirToScan)
    if not res['OK']:
      return res
    subDirs = res['Value']['Successful'][dirToScan]['SubDirs']
    doneDirs = set()
    # Get summary for each next level directory in order to reduce the size
    for directory in subDirs:
      # For  top directories, go one level below to reduce the size of information
      if len(directory.split(os.path.sep)) == 3:
        self.__getTopDirUsage(directory, topDirLogicalUsage)
        topDir = _standardDirectory(directory)
      elif directory.lower() not in doneDirs:
        # The StorageUsage DB is case insensitive, hence go only once over the same directory
        doneDirs.add(directory.lower())
        directory = _standardDirectory(directory)
        # get info from the DB about the LOGICAL STORAGE USAGE (from the su_Directory table):
        result = self.__stDB.getSummary(directory)
        if not result['OK']:
          return result
        logicalUsage = result['Value']
        # Store logical usage for top level directory onl
        topDir = _standardDirectory(os.path.join(os.path.sep, *directory.split(os.path.sep)[:3]))
        for row in logicalUsage:
          # d, size, files = row
          topDirLogicalUsage[topDir]['Files'] += logicalUsage[row]['Files']
          topDirLogicalUsage[topDir]['Size'] += logicalUsage[row]['Size']
      self.log.verbose("After scan of %s, total of %s: " % (directory, topDir),
                       "size: %.4f TB  files: %d" % (convertSizeUnits(topDirLogicalUsage[topDir]['Size'], 'B', 'TB'),
                                                     topDirLogicalUsage[topDir]['Files']))
    return S_OK()

  def topDirectoryAccounting(self):
    """
    Get statistics for top level 2 directories, e.g. /lhcb/data/, /lhcb/user/ etc...
    """
    self.log.notice("-------------------------------------------------------------------------------------\n")
    self.log.notice("Generate accounting records for top directories ")
    self.log.notice("-------------------------------------------------------------------------------------\n")

    # Get second level top directories
    topDirLogicalUsage = defaultdict(lambda: {'Files': 0, 'Size': 0})  # build the list of first level directories

    res = self.__getTopDirUsage('/lhcb', topDirLogicalUsage)
    if not res['OK']:
      self.log.error("Error getting top directory logical usage", res['Message'])
      return res
    self.log.notice("Summary on logical usage of top directories: ")
    for topDir in topDirLogicalUsage:
      self.log.notice("dir: %s size: %.4f TB  files: %d" %
                      (topDir,
                       convertSizeUnits(topDirLogicalUsage[topDir]['Size'], 'B', 'TB'),
                       topDirLogicalUsage[topDir]['Files']))

    # loop on top level directories (/lhcb/data/, /lhcb/user/, /lhcb/MC/, etc..)
    # to get the summary in terms of PHYSICAL usage grouped by SE:
    seData = {}
    for directory in topDirLogicalUsage:
      result = self.__stDB.getDirectorySummaryPerSE(directory)  # retrieve the PHYSICAL usage
      if not result['OK']:
        return result
      seData[directory] = result['Value']
      self.log.notice("Got SE summary for %s directories " % (len(seData)))
      self.log.debug("SEData: %s" % seData)
    # loop on top level directories to send the accounting records
    numRows = 0
    now = datetime.datetime.utcnow()
    for directory in seData:
      self.log.debug("dir: %s SEData: %s " % (directory, seData[directory]))
      if directory not in topDirLogicalUsage:
        self.log.error("Dir %s is in the summary per SE, but it is not in the logical files summary!" % directory)
        continue
      for se in sorted(seData[directory]):
        storageRecord = Storage()
        storageRecord.setStartTime(now)
        storageRecord.setEndTime(now)
        storageRecord.setValueByKey("Directory", directory)
        storageRecord.setValueByKey("StorageElement", se)
        storageRecord.setValueByKey("LogicalFiles", topDirLogicalUsage[directory]['Files'])
        storageRecord.setValueByKey("LogicalSize", topDirLogicalUsage[directory]['Size'])
        try:
          physicalFiles = seData[directory][se]['Files']
        except Exception:
          self.log.error("WARNING! no files replicas for directory %s on SE %s" % (directory, se))
          physicalFiles = 0
        try:
          physicalSize = seData[directory][se]['Size']
        except Exception:
          self.log.error("WARNING! no size for replicas for directory %s on SE %s" % (directory, se))
          physicalSize = 0
        storageRecord.setValueByKey("PhysicalFiles", physicalFiles)
        storageRecord.setValueByKey("PhysicalSize", physicalSize)
        gDataStoreClient.addRegister(storageRecord)
        numRows += 1
        self.log.debug("Directory: %s SE: %s  physical size: %.4f TB (%d files)" %
                       (directory, se,
                        convertSizeUnits(physicalSize, 'B', 'TB'),
                        physicalFiles))

    self.log.notice("Sending %s records to accounting for top level directories storage" % numRows)
    res = gDataStoreClient.commit()
    if not res['OK']:
      self.log.notice("ERROR: committing Storage records: %s " % res)
      return S_ERROR(res)
    else:
      self.log.notice("%s records for Storage type successfully committed" % numRows)

  def bkPathAccounting(self):
    """
    Generate accounting for all directories, grouped by BK path metadata
    """
    self.log.notice("-------------------------------------------------------------------------------------\n")
    self.log.notice("Generate accounting records for DataStorage type ")
    self.log.notice("-------------------------------------------------------------------------------------\n")

    # counter for DataStorage records, commit to the accounting in bunches of self.limitForCommit records
    self.totalRecords = 0
    self.recordsToCommit = 0
    self.log.notice(" Call the function to extract information from the StorageUsageDB..")
    res = self.generateStorageUsagePerDir()
    if not res['OK']:
      self.log.error("ERROR querying the StorageUsageDB per directory")
      return S_ERROR()

    # Keep a list of all directories in FC that are not found in the Bkk
    self.directoriesNotInBkk = []
    # for debugging purposes build dictionaries with storage usage to compare with the accounting plots
    self.debug_seUsage = defaultdict(lambda: {'Files': 0, 'Size': 0})
    self.debug_seUsage_acc = defaultdict(lambda: {'Files': 0, 'Size': 0})

    # set the time for the accounting records (same time for all records)
    now = datetime.datetime.utcnow()
    # Get the directory metadata in a bulk query
    metaForList = self.__getMetadataForAcc(self.dirDict.values())

    # loop on all directories  to get the bkk metadata
    for dirLfn, fullDirectory in self.dirDict.items():
      if dirLfn not in fullDirectory:
        self.log.error("ERROR: fullDirectory should include the dirname: %s %s " % (fullDirectory, dirLfn))
        continue
      self.log.verbose("Processing directory %s " % dirLfn)
      if dirLfn not in self.pfnUsage:
        self.log.error("ERROR: directory does not have PFN usage %s " % dirLfn)
        continue
      self.log.verbose("PFN usage: %s " % self.pfnUsage[dirLfn])
      if dirLfn not in self.lfnUsage:
        self.log.error("ERROR: directory does not have LFN usage %s " % dirLfn)
        continue
      self.log.verbose("LFN usage: %s " % self.lfnUsage[dirLfn])

      # for DEBUGGING:
      for se in self.pfnUsage[dirLfn]:
        self.debug_seUsage[se]['Files'] += self.pfnUsage[dirLfn][se]['Files']
        self.debug_seUsage[se]['Size'] += self.pfnUsage[dirLfn][se]['Size']
      # end of DEBUGGING

      # get metadata for this directory
      metaForDir = metaForList.get(fullDirectory, {})
      if not metaForDir:
        self.log.warn("Metadata not available for directory %s" % fullDirectory)
        continue

      # Fill in the accounting record
      self.log.info("Fill the record for %s and metadata: %s " % (dirLfn, metaForDir))
      res = self.fillAndSendAccountingRecord(dirLfn, metaForDir, now)
      if not res['OK']:
        return res
      for se in self.pfnUsage[dirLfn]:
        self.debug_seUsage_acc[se]['Files'] += self.pfnUsage[dirLfn][se]['Files']
        self.debug_seUsage_acc[se]['Size'] += self.pfnUsage[dirLfn][se]['Size']

    # Don't forget to commit the remaining records!
    self.__commitRecords()

  def execute(self):
    if self.am_getOption("CleanBefore", False):
      self.log.notice("Cleaning the DB")
      result = self.__stDB.purgeOutdatedEntries("/lhcb/user", self.am_getOption("OutdatedSeconds", 86400 * 10))
      if not result['OK']:
        return result
      self.log.notice("Purged %s outdated records" % result['Value'])

    # User accounting (per user and SE)
    self.userStorageAccounting()
    # Accounting per top directory
    self.topDirectoryAccounting()
    # full production data accounting
    self.bkPathAccounting()

    self.log.notice("-------------------------------------------------------------------------------------\n")
    self.log.notice("------ End of cycle report for DataStorage accounting--------------------------------\n")
    self.log.notice("Total directories found in FC:  %d " % len(self.dirDict))
    totalCallsToStorageUsage = self.callsToGetSummary + self.callsToDirectorySummary
    self.log.notice("Total calls to StorageUsage: %d , took: %d s " % (totalCallsToStorageUsage, self.genTotalTime))
    totalCallsToBkk = self.callsToBkkgetDirectoryMetadata + self.callsToBkkGetEvtType
    self.log.notice("Total calls to DataUsage for cache: %d" % self.callsToDudbForMetadata)
    self.log.notice("Total calls to Bookkeeping: %d (getDirectoryMetadata: %d, getEventType: %d)" %
                    (totalCallsToBkk, self.callsToBkkgetDirectoryMetadata, self.callsToBkkGetEvtType))
    self.log.notice("Total records sent to accounting for DataStorage:  %d " % self.totalRecords)
    self.log.notice("Directories not found in Bookkeeping: %d " % (len(self.directoriesNotInBkk)))
    fileName = os.path.join(self.__workDirectory, "directoriesNotInBkk.txt")
    self.log.notice("written to file: %s " % fileName)
    fd = open(fileName, "w")
    for dd in self.directoriesNotInBkk:
      fd.write("%s\n" % dd)
    fd.close()
    # for DEBUG only
    self.log.info("Summary of StorageUsage: files size ")
    for se in sorted(self.debug_seUsage):
      self.log.info("all: %s  %d %d Bytes ( %.2f TB ) " % (se, self.debug_seUsage[se]['Files'],
                                                           self.debug_seUsage[se]['Size'],
                                                           self.debug_seUsage[se]['Size'] / 1.0e12))
      if se in self.debug_seUsage_acc:
        self.log.info("acc: %s  %d %d Bytes ( %.2f TB ) " % (se, self.debug_seUsage_acc[se]['Files'],
                                                             self.debug_seUsage_acc[se]['Size'],
                                                             self.debug_seUsage_acc[se]['Size'] / 1.0e12))
      else:
        self.log.info("SE not in self.debug_seUsage_acc keys")
    return S_OK()

  def __getMetadataForAcc(self, dirList):
    """Get metadata for a directory either from memory, from the storageDB or
    from BK."""
    # Try and get the metadata from memory cache
    notFound = []
    metaForList = {}
    for dirName in dirList:
      metaForList[dirName] = self.cachedMetadata.get(dirName, {})
      if not metaForList[dirName]:
        notFound.append(dirName)
    notInCache = []
    if notFound:
      self.log.info("Memory metadata cache missed for %d directories" % len(notFound))
      self.log.verbose("call getDirMetadata for (first 10): %s " % str(notFound[0:10]))
      for dirChunk in breakListIntoChunks(notFound, 10000):
        self.callsToDudbForMetadata += 1
        res = self.__dataUsageClient.getDirMetadata(dirChunk)  # this could be a bulk query for a list of directories
        if not res['OK']:
          self.log.error("Error retrieving %d directories meta-data %s " % (len(dirChunk), res['Message']))
          # this usually happens when directories are removed from FC between the StorageUsageDB dump and this call,
          # if the Db is refreshed exactly in this time interval. Not really a problem.
          # 3 just a try ##############################################3
          notInCache += dirChunk
          continue
        self.log.verbose("getDirMetadata returned: %s " % str(res['Value']))
        for dirName in dirChunk:
          # Compatibility with old (list for single file) and new (dictionary) service
          if isinstance(res['Value'], type({})):
            metaTuple = res['Value'].get(dirName, ())
          elif len(dirList) == 1 and res['Value']:
            metaTuple = res['Value'][0]
          else:
            metaTuple = ()
          if metaTuple and metaTuple[3] is not None:
            metaForDir = metaForList[dirName]
            _dirID, metaForDir['DataType'], metaForDir['Activity'], \
                metaForDir['Conditions'], metaForDir['ProcessingPass'], \
                metaForDir['EventType'], metaForDir['FileType'], \
                metaForDir['Production'], metaForDir['Visibility'] = metaTuple
          else:
            notInCache.append(dirName)

      failedBK = []
      if notInCache:
        cachedFromBK = []
        self.log.info("Directory metadata cache missed for %d directories => query BK and cache" % len(notInCache))
        for dirChunk in breakListIntoChunks(notInCache, 200):
          self.callsToBkkgetDirectoryMetadata += 1
          res = self.__bkClient.getDirectoryMetadata(dirChunk)
          if not res['OK']:
            self.log.error("Totally failed to query Bookkeeping", res['Message'])
            failedBK += dirChunk
            for dirName in dirChunk:
              metaForDir = metaForList[dirName]
              _fillMetadata(metaForDir, 'FailedBkkQuery')
          else:
            bkMetadata = res['Value']
            self.log.debug("Successfully queried Bookkeeping, result: %s " % bkMetadata)
            for dirName in dirChunk:
              metaForDir = metaForList[dirName]
              # BK returns a list of metadata, chose the first one...
              metadata = bkMetadata['Successful'].get(dirName, [{}])[0]
              if metadata and metadata.get('ConditionDescription') is not None:
                metadata['Visibility'] = metadata.pop('VisibilityFlag', metadata.get('Visibility', 'na'))
                # All is OK, directory found
                _fillMetadata(metaForDir, metadata)
                self.log.verbose("Cache entry %s in DirMetadata table.." % dirName)
                resInsert = self.__dataUsageClient.insertToDirMetadata({dirName: metadata})
                if not resInsert['OK']:
                  self.log.error("Failed to cache metadata:", "%s for dir %s" % (resInsert['Message'], dirName))
                else:
                  cachedFromBK.append(dirName)
                  self.log.verbose("Successfully cached metadata for %s : %s" % (dirName, str(metadata)))
                  self.log.debug("result: %s " % str(resInsert))
              else:
                # Directory not found
                self.log.verbose("Directory %s not registered in Bookkeeping!" % dirName)
                _fillMetadata(metaForDir, 'notInBkk')
                failedBK.append(dirName)
                self.directoriesNotInBkk.append(dirName)
              # Translate a few keys for accounting
              for bkName, accName in (('ConfigName', 'DataType'), ('ConfigVersion', 'Activity'),
                                      ('ConditionDescription', 'Conditions')):
                metaForDir[accName] = metaForDir.pop(bkName, 'na')
        self.log.info('Successfully cached %d directories from BK' % len(cachedFromBK))
        if self.directoriesNotInBkk:
          self.log.warn('%d directories not found in BK' % len(self.directoriesNotInBkk))

      # cache locally the metadata
      for dirName in [dn for dn in notFound if dn not in failedBK]:
        metaForDir = metaForList[dirName]
        # Translate the numerical event type to a description string
        metaForDir['EventType'] = self.__getEventTypeDescription(metaForDir.pop('EventType', 'na'))
        self.cachedMetadata[dirName] = metaForDir.copy()
    else:
      self.log.info("Memory metadata cache hit for %d directories" % len(dirList))
    return metaForList

  def __commitRecords(self):
    if self.recordsToCommit:
      res = gDataStoreClient.commit()
      if not res['OK']:
        self.log.error("Accounting ERROR: commit returned %s" % res)
      else:
        self.log.notice("%d records committed " % self.recordsToCommit)
        self.recordsToCommit = 0
        self.log.notice("commit for DataStorage returned: %s" % res)

  def generateStorageUsagePerDir(self):
    """
    Extract storage info from the StorageUsageDB and keep it in memory in
    dictionaries self.lfnUsage and self.pfnUsage
    """

    start = time.time()
    self.log.notice('Starting from path: /lhcb/')
    res = self.__stDB.getStorageDirectories('/lhcb/')
    if not res['OK']:
      return S_ERROR()
    totalDirList = res['Value']
    self.log.info("Total directories retrieved from StorageUsageDB: %d " % len(totalDirList))
    # select only terminal directories (directories without sub-directories)
    # mc directory structure: /lhcb/MC/[year]/[file type]/[prod]/0000/ => len = 7
    # raw data:               /lhcb/data/2011/RAW/FULL/LHCb/COLLISION11/99983
    # => len 9 (under /lhcb/data/ from 2011 only RAW, before 2011 also other file types)
    # processed data: under both /lhcb/data and /lhcb/LHCb/
    #                         /lhcb/data/2010/DST/00009300/0000
    # data:                   /lhcb/LHCb/Collision12/ICHEP.DST/00017399/0000/
    self.dirDict = {}
    ignoredDirectories = dict.fromkeys(self.__ignoreDirsList, 0)
    self.log.info("Directories to be ignored: %s " % str(sorted(ignoredDirectories)))
    for dirItem in totalDirList:
      # make sure that last character is a '/'
      dirItem = _standardDirectory(dirItem)
      splitDir = dirItem.split(os.path.sep)
      if len(splitDir) < 4:  # avoid picking up intermediate directories which don't contain files, like /lhcb/
        self.log.warn("Directory %s skipped, as top directory" % dirItem)
        continue
      secDir = splitDir[2]
      if secDir in ignoredDirectories:
        self.log.verbose("Directory to be ignored, skipped: %s " % dirItem)
        ignoredDirectories[secDir] += 1
        continue
      # for each type of directory (MC, reconstructed data and raw data) check the format,
      # in order not to count more than once the productions with more than one sub-directory
      # for MC directories:
      # example: '/lhcb/MC/MC10/ALLSTREAMS.DST/00010908/0000/',
      # or        /lhcb/MC/2011/DST/00010870/0000
      # one directory for each file type
      # for histograms, there is no numeric "suffix" like /0000/
      # example: /lhcb/LHCb/Ionproton13/HIST/136973/
      # for data
      # production: /lhcb/LHCb/Collision11/SWIMSTRIPPINGD02KSPIPI.MDST/00019088/0000/
      # for raw data: /lhcb/data/2012/RAW/FULL/LHCb/COLLISION12/133784/
      try:
        # RAW data directories have a special format, see above
        if splitDir[-6] == "RAW":
          self.log.verbose("RAW DATA directory: %s" % splitDir)
          directory = dirItem
          fullDirectory = directory
        else:
          # These are production directories, see above for the format
          self.log.verbose("MC or reconstructed data directory: %s" % splitDir)
          # HIST directories do not have a "suffix" (i.e. all in the same directory)!
          if splitDir[-3] == 'HIST':
            directory = dirItem
            fullDirectory = directory
          else:
            # Ignore the suffix for the accounting as this is only an artifact
            directory = _standardDirectory(os.path.dirname(os.path.dirname(dirItem)))
            fullDirectory = dirItem
        # Keep the link between the accounting directory and the full directory name
        self.dirDict.setdefault(directory, fullDirectory)
        self.log.verbose("Directory contains production files: %s " % directory)
      except Exception:
        self.log.warn("The directory has unexpected format: %s " % dirItem)

    self.lfnUsage = defaultdict(lambda: {'LfnSize': 0, 'LfnFiles': 0})
    self.pfnUsage = {}

    totalDiscardedDirs = 0
    self.log.info("Directories that have been discarded:")
    for dd in ignoredDirectories:
      self.log.info("/lhcb/%s - %d " % (dd, ignoredDirectories[dd]))
      totalDiscardedDirs += ignoredDirectories[dd]
    self.log.info("Total discarded directories: %d " % totalDiscardedDirs)

    self.log.info("Retrieved %d dirs from StorageUsageDB containing prod files" % len(self.dirDict))
    self.log.info("Getting the number of files and size from StorageUsage service")

    for directory in self.dirDict:
      self.log.verbose("Get storage usage for directory %s " % directory)
      res = self.__stDB.getDirectorySummaryPerSE(directory)
      self.callsToDirectorySummary += 1
      if not res['OK']:
        self.log.error("Cannot retrieve PFN usage", res['Message'])
        continue
      # save the PFN usage per SE in a dictionary
      self.pfnUsage.setdefault(directory, res['Value'])

      self.log.verbose("Get logical usage for directory %s " % directory)
      # This returns the number of files and size for the directory and all its subdirectories
      res = self.__stDB.getSummary(directory)
      self.callsToGetSummary += 1
      if not res['OK']:
        self.log.error("Cannot retrieve LFN usage", res['Message'])
        continue
      if not res['Value']:
        self.log.error("getSummary returned empty value", "for %d: %s" % (directory, str(res)))
        continue
      # Sum up all subdirectories
      for dirInfo in res['Value'].values():  # can be an iterator
        self.lfnUsage[directory]['LfnSize'] += dirInfo['Size']
        self.lfnUsage[directory]['LfnFiles'] += dirInfo['Files']
      self.log.verbose("PFN usage: %s" % self.pfnUsage[directory])
      self.log.verbose("LFN usage: %s" % self.lfnUsage[directory])

    end = time.time()
    self.genTotalTime = end - start
    self.log.info("StorageUsageDB extraction completed in %d s" % self.genTotalTime)

    return S_OK()

  def __getEventTypeDescription(self, eventType):
    # convert eventType to string:
    try:
      eventType = int(eventType)
    except Exception:
      pass
    # check that the event type description is in the cached dictionary, and otherwise query the Bkk
    if eventType not in self.eventTypeDescription:
      self.log.notice("Event type description not available for eventTypeID %s, getting from Bkk" % eventType)
      res = self.__bkClient.getAvailableEventTypes()
      self.callsToBkkGetEvtType += 1
      if not res['OK']:
        self.log.error("Error querying the Bkk:", res['Message'])
      else:
        self.eventTypeDescription.update(dict(res['Value']))
      self.log.verbose("Updated  self.eventTypeDescription dict: %s " % str(self.eventTypeDescription))
      # If still not found, log it!
      if eventType not in self.eventTypeDescription:
        self.log.error("EventType %s is not in cached dictionary" % str(eventType))

    return self.eventTypeDescription.get(eventType, 'na')

  def fillAndSendAccountingRecord(self, lfnDir, metadataDict, now):
    """Create, fill and send to accounting a record for the DataStorage
    type."""
    dataRecord = DataStorage()
    dataRecord.setStartTime(now)
    dataRecord.setEndTime(now)
    logicalSize = self.lfnUsage[lfnDir]['LfnSize']
    logicalFiles = self.lfnUsage[lfnDir]['LfnFiles']
    dataRecord.setValueByKey("LogicalSize", logicalSize)
    dataRecord.setValueByKey("LogicalFiles", logicalFiles)
    for key in ('DataType', 'Activity', 'FileType', 'Production', 'ProcessingPass', 'Conditions', 'EventType'):
      dataRecord.setValueByKey(key, metadataDict.get(key, 'na'))
    self.log.verbose(">>> Send DataStorage record to accounting:")
    self.log.verbose("\tlfnFiles: %d lfnSize: %d " % (logicalFiles, logicalSize))

    for se in self.pfnUsage[lfnDir]:
      self.log.verbose("Filling accounting record for se %s" % se)
      physicalSize = self.pfnUsage[lfnDir][se]['Size']
      physicalFiles = self.pfnUsage[lfnDir][se]['Files']

      dataRecord.setValueByKey("StorageElement", se)
      dataRecord.setValueByKey("PhysicalSize", physicalSize)
      dataRecord.setValueByKey("PhysicalFiles", physicalFiles)
      self.log.verbose("\t\tStorageElement: %s --> physFiles: %d  physSize: %d " % (se, physicalFiles, physicalSize))

      # addRegister is NOT making a copy, therefore all records are otherwise overwritten
      res = gDataStoreClient.addRegister(copy.deepcopy(dataRecord))
      if not res['OK']:
        self.log.error("addRegister returned: %s" % res)
        return S_ERROR("addRegister returned: %s" % res)
      # Reset logical information to zero in order to send it only once!
      dataRecord.setValueByKey("LogicalSize", 0)
      dataRecord.setValueByKey("LogicalFiles", 0)
      self.totalRecords += 1
      self.recordsToCommit += 1

    # Commit if necessary
    if self.recordsToCommit > self.limitForCommit:
      self.__commitRecords()

    return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""just some common components for tests."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from mock import MagicMock

from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

version = 'someVers'
prod_id = '123'
prod_job_id = '00000456'
wms_job_id = 12345
workflowStatus = {'OK': True}
stepStatus = {'OK': True}
step_number = '321'
step_id = '%s_%s_%s' % (prod_id, prod_job_id, step_number)

ar_mock = MagicMock()
ar_mock.commit.return_value = {'OK': True, 'Value': ''}

jr_mock = MagicMock()
jr_mock.setApplicationStatus.return_value = {'OK': True, 'Value': ''}
jr_mock.generateForwardDISET.return_value = {'OK': True, 'Value': Operation()}
jr_mock.setJobParameter.return_value = {'OK': True, 'Value': 'pippo'}

fr_mock = MagicMock()
fr_mock.getFiles.return_value = {}
fr_mock.setFileStatus.return_value = {'OK': True, 'Value': ''}
fr_mock.commit.return_value = {'OK': True, 'Value': ''}
fr_mock.generateForwardDISET.return_value = {'OK': True, 'Value': Operation()}

rc_mock = Request()
rc_mock.RequestName = 'aRequestName'
rc_mock.OwnerDN = 'pippo'
rc_mock.OwnerGroup = 'pippoGRP'
rOp = Operation()
rOp.Type = 'PutAndRegister'
rOp.TargetSE = 'anSE'
f = File()
f.LFN = '/foo/bar.py'
f.PFN = '/foo/bar.py'
rOp.addFile(f)
rc_mock.addOperation(rOp)


wf_commons = [{'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id, 'eventType': '123456789', 'jobType': 'merge',
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion', 'outputDataFileMask': '',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData', 'numberOfEvents': '100',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'runNumber': 'Unknown', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'merge',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData', 'numberOfEvents': '100',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'runNumber': 'Unknown',
               'gaudiSteps': ['someApp_1'], 'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'merge',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData', 'numberOfEvents': '100',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'merge',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData', 'numberOfEvents': '100',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'runNumber': 'Unknown', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'runNumber': 'Unknown',
               'gaudiSteps': ['someApp_1'], 'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'InputData': '', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'InputData': 'foo;bar', 'gaudiSteps': ['someApp_1'],
               'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'InputData': 'foo;bar', 'ParametricInputData': '',
               'gaudiSteps': ['someApp_1'], 'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}},
              {'PRODUCTION_ID': prod_id, 'JOB_ID': prod_job_id,
               'configName': 'aConfigName', 'configVersion': 'aConfigVersion',
               'outputDataFileMask': '', 'jobType': 'reco',
               'BookkeepingLFNs': 'aa', 'ProductionOutputData': 'ProductionOutputData',
               'JobReport': jr_mock, 'Request': rc_mock, 'AccountingReport': ar_mock, 'FileReport': fr_mock,
               'SystemConfig': 'sys_config', 'LogFilePath': 'someDir', 'LogTargetPath': 'someOtherDir',
               'runNumber': 'Unknown', 'InputData': 'foo;bar', 'ParametricInputData': 'pid1;pid2;pid3',
               'gaudiSteps': ['someApp_1'], 'outputSEs':{"DAVINCIHIST": "CERN-HIST", "TXT": "SE1"}}, ]

step_commons = [{'applicationName': 'someApp',
                 'applicationVersion': 'v1r0',
                 'eventType': '123456789',
                 'SystemConfig': 'sys_config',
                 'applicationLog': 'appLog',
                 'extraPackages': '',
                 'XMLSummary': 'XMLSummaryFile',
                 'numberOfEvents': '100',
                 'BKStepID': '123',
                 'StepProcPass': 'Sim123',
                 'outputFilePrefix': 'pref_',
                 'STEP_INSTANCE_NAME': 'someApp_1',
                 'inputData': '/for/bar/',
                 'listoutput': [{'outputDataName': prod_id + '_' + prod_job_id + '_',
                                 'outputDataType': 'bbb'}]},
                {'applicationName': 'someApp',
                 'applicationVersion': 'v1r0',
                 'eventType': '123456789',
                 'SystemConfig': 'sys_config',
                 'applicationLog': 'appLog',
                 'extraPackages': '',
                 'XMLSummary': 'XMLSummaryFile',
                 'numberOfEvents': '100',
                 'BKStepID': '123',
                 'StepProcPass': 'Sim123',
                 'outputFilePrefix': 'pref_',
                 'optionsLine': '',
                 'inputData': '/for/bar/',
                 'STEP_INSTANCE_NAME': 'someApp_1',
                 'listoutput': [{'outputDataName': prod_id + '_' + prod_job_id + '_',
                                 'outputDataType': 'bbb'}]},
                {'applicationName': 'someApp',
                 'applicationVersion': 'v1r0',
                 'eventType': '123456789',
                 'SystemConfig': 'sys_config',
                 'applicationLog': 'appLog',
                 'extraPackages': '',
                 'XMLSummary': 'XMLSummaryFile',
                 'numberOfEvents': '100',
                 'BKStepID': '123',
                 'StepProcPass': 'Sim123',
                 'outputFilePrefix': 'pref_',
                 'extraOptionsLine': 'blaBla',
                 'inputData': '/for/bar/',
                 'STEP_INSTANCE_NAME': 'someApp_1',
                 'listoutput': [{'outputDataName': prod_id + '_' + prod_job_id + '_',
                                 'outputDataType': 'bbb'}]}]



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client for BookkeepingDB file catalog."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

__RCSID__ = "$Id$"


class BookkeepingDBClient(FileCatalogClientBase):
  """File catalog client for bookkeeping DB."""

  READ_METHODS = FileCatalogClientBase.READ_METHODS + ['isDirectory', 'isLink', 'getFileSize', 'getFileMetadata']
  WRITE_METHODS = FileCatalogClientBase.WRITE_METHODS + ['addFile',
                                                         'addReplica',
                                                         'removeFile',
                                                         'removeReplica',
                                                         'setReplicaStatus',
                                                         'setReplicaProblematic',
                                                         'setReplicaHost',
                                                         'removeDirectory',
                                                         'createDirectory',
                                                         'removeLink',
                                                         'createLink']

  def __init__(self, **kwargs):
    """Constructor of the Bookkeeping catalogue client."""
    self.splitSize = 1000
    self.name = 'BookkeepingDB'
    self.server = BookkeepingClient(timeout=120)

  @checkCatalogArguments
  def addFile(self, lfn):
    """Set the replica flag."""
    return self.__setHasReplicaFlag(lfn)

  @checkCatalogArguments
  def addReplica(self, lfn):
    """Same as addFile."""
    return self.addFile(lfn)

  @checkCatalogArguments
  def removeFile(self, path):
    """Remove the replica flag."""
    return self.__unsetHasReplicaFlag(path)

  @checkCatalogArguments
  def isFile(self, lfn):
    """Returns a dictionary True/False."""
    return self.__exists(lfn)

  @checkCatalogArguments
  def isDirectory(self, lfn):
    """ Return Successful dict: True if lfn is a directory, False if a file - Failed dict if not existing
    """
    res = self.isFile(lfn)
    if not res['OK']:
      successful = {}
      failed = dict.fromkeys(lfn, res['Message'])
    else:
      successful = dict.fromkeys([lfn for lfn, val in res['Value']['Successful'].items() if val], False)
      failed = res['Value']['Failed']
      toCheck = [lfn for lfn, val in res['Value']['Successful'].items() if not val]
      if toCheck:
        # Can't use service directly as
        res = self.server.getDirectoryMetadata_new(toCheck)
        if not res['OK']:
          failed.update(dict.fromkeys(toCheck, res['Message']))
        else:
          failed.update(dict.fromkeys([lfn for lfn in res['Value']['Failed']], 'Not a file or directory'))
          successful.update(dict.fromkeys([lfn for lfn in res['Value']['Successful']], True))
    return S_OK({'Successful': successful, 'Failed': failed})

  @checkCatalogArguments
  def isLink(self, lfn):
    return self.__returnSuccess(lfn, val=False)

  @checkCatalogArguments
  def __returnSuccess(self, lfn, val=True):
    """Generic method returning success for all input files."""
    return S_OK({'Failed': {}, 'Successful': dict.fromkeys(lfn, val)})

  @checkCatalogArguments
  def removeReplica(self, lfn):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def setReplicaStatus(self, lfn):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def setReplicaProblematic(self, lfn, revert=False):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def setReplicaHost(self, lfn):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def removeDirectory(self, lfn, recursive=False):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def createDirectory(self, lfn):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def removeLink(self, lfn):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def createLink(self, lfn):
    return self.__returnSuccess(lfn)

  @checkCatalogArguments
  def exists(self, path):
    """Returns a dictionary of True/False on file existence."""
    return self.__exists(path)

  @checkCatalogArguments
  def getFileMetadata(self, path):
    """Return the metadata dictionary."""
    return self.__getFileMetadata(path)

  @checkCatalogArguments
  def getFileSize(self, path):
    """Return just the file size."""

    res = self.__getFileMetadata(path)
    # Always returns OK
    successful = dict([(lfn, metadata['FileSize']) for lfn, metadata in res['Value']['Successful'].items()])
    return S_OK({'Successful': successful, 'Failed': res['Value']['Failed']})

  ################################################################
  #
  # These are the internal methods used for actual interaction with the BK service
  #

  def __checkArgumentFormat(self, path):
    """Returns a list, either from a string or keys of a dict."""
    if isinstance(path, six.string_types):
      return S_OK([path])
    elif isinstance(path, list):
      return S_OK(path)
    elif isinstance(path, dict):
      return S_OK(list(path))
    else:
      errStr = "BookkeepingDBClient.__checkArgumentFormat: Supplied path is not of the correct format."
      gLogger.error(errStr)
      return S_ERROR(errStr)

  def __toggleReplicaFlag(self, lfns, setflag=True):
    successful = {}
    # Poor man's way to not return an error for user files
    for lfn in [lfn for lfn in lfns if lfn.startswith('/lhcb/user')]:
      lfns.pop(lfn)
      successful[lfn] = True
    failed = {}
    for lfnList in breakListIntoChunks(lfns, self.splitSize):
      res = {True: self.server.addFiles, False: self.server.removeFiles}[setflag](lfnList)
      if not res['OK']:
        failed.update(dict.fromkeys(lfnList, res['Message']))
      else:
        # It is dirty, but ...
        failed.update(dict.fromkeys([lfn for lfn in res['Value']['Failed']], 'File does not exist'))
        successful.update(dict.fromkeys([lfn for lfn in res['Value']['Successful']], True))
    return S_OK({'Successful': successful, 'Failed': failed})

  def __setHasReplicaFlag(self, lfns):
    """Set replica flags on BKK."""
    return self.__toggleReplicaFlag(lfns, setflag=True)

  def __unsetHasReplicaFlag(self, lfns):
    """Removes replica flags on BKK."""
    return self.__toggleReplicaFlag(lfns, setflag=False)

  def __exists(self, lfns):
    """Checks if lfns exist."""
    successful = {}
    failed = {}
    for lfnList in breakListIntoChunks(lfns, self.splitSize):
      res = self.server.exists(lfnList)
      if not res['OK']:
        failed.update(dict.fromkeys(lfnList, res['Message']))
      else:
        successful.update(res['Value'])
    return S_OK({'Successful': successful, 'Failed': failed})

  @checkCatalogArguments
  def __getFileMetadata(self, lfns):
    """Returns lfns metadata."""
    successful = {}
    failed = {}
    for lfnList in breakListIntoChunks(lfns, self.splitSize):
      res = self.server.getFileMetadata(lfnList)
      if not res['OK']:
        failed.update(dict.fromkeys(lfnList, res['Message']))
      else:
        success = res['Value'].get('Successful', res['Value'])
        failed.update(dict.fromkeys((lfn for lfn in lfnList if lfn not in success), 'File does not exist'))
        failed.update(dict((lfn, val) for lfn, val in success.items() if isinstance(val, six.string_types)))
        successful.update(dict((lfn, val) for lfn, val in success.items() if not isinstance(val, six.string_types)))
    return S_OK({'Successful': successful, 'Failed': failed})



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Methods for scanning the popularity table."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import time
from datetime import datetime, timedelta

import DIRAC
from DIRAC import gLogger
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup

from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient
from LHCbDIRAC.DataManagementSystem.Client.DataUsageClient import DataUsageClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery, makeBKPath
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

# Code
dmsHelper = DMSHelpers()

# Dictionary containing the BK path for each directory
bkPathForDir = {}
# Set of invisible BK paths
cachedInvisible = set()
# Set of productions for each BK path
prodForBKPath = {}
# Storage usage for each BK path
bkPathUsage = {}
# Processing pass for each BK path
processingPass = {}
# PFN #files and size for each BK path
physicalStorageUsage = {}

# Dictionary with weekly/dayly usage for each BK path
bkPathPopularity = {}

storageTypes = ('Disk', 'Tape', 'Archived', 'All', 'LFN')
# Set of BK paths for each storage types (above)
datasetStorage = {}

# We are only interested in CERN and Tier1s
storageSites = dmsHelper.getTiers(tier=(0, 1))
# Caching the site at which an SE is
cachedSESites = {}

# Instantiate clients
duClient = DataUsageClient()
suClient = StorageUsageClient()
bkClient = BookkeepingClient()
transClient = TransformationClient()


def getTimeBin(date):
  """get the bin number as week number."""
  year, week, _day = date.isocalendar()
  week += 52 * year
  return week


def cacheDirectories(directories):
  """Get directories information from popularity and from storageUsage and
  cache the information."""

  # Ignore already cached directories
  dirSet = directories - set(bkPathForDir) - cachedInvisible
  if not dirSet:
    return
  startTime = time.time()
  gLogger.always('\nCaching %d directories, %d not cached yet' % (len(directories), len(dirSet)))

  # Some DBs refer to "long" directory names (i.e. with additional number, see below)
  #     while others refer to the unnumbered directory.
  # Make conversion tables and keep only the first directory when numbered
  # Short is for example /lhcb/LHCb/Collision18/RDST/00074101/
  # Long is for example /lhcb/LHCb/Collision18/RDST/00074101/0008/
  dirLong2Short = {}
  dirShort2Long = {}
  for dirLfn in sorted(dirSet):
    splitDir = dirLfn.split('/')
    if len(splitDir[-2]) == 4 and splitDir[-2].isdigit():
      short = '/'.join(splitDir[:-2])
    else:
      short = dirLfn
    dirLong2Short[dirLfn] = short
    dirShort2Long.setdefault(short, dirLfn)

  # # Cache BK path for directories
  # First try cached information in Data Usage DB
  chunkSize = 1000
  gLogger.always('Getting cache metadata for %d directories' % len(dirShort2Long))
  dirMetadata = {}
  for longDirs in breakListIntoChunks(dirShort2Long.values(), chunkSize):
    res = duClient.getDirMetadata(longDirs)
    if not res['OK']:
      gLogger.error("\nError getting metadata from DataUsage", res['Message'])
    else:
      dirMetadata.update(res['Value'])

  invisible = set()
  lfnsFromBK = set()
  for dirLfn in dirSet:
    longDir = dirShort2Long[dirLong2Short[dirLfn]]
    metadata = dirMetadata.get(longDir)
    if metadata:
      if len(metadata) < 9 or metadata[8] != "N":
        bkPathForDir[dirLfn] = os.path.join('/',
                                            metadata[1], metadata[2],
                                            metadata[3], metadata[4][1:],
                                            metadata[5], metadata[6])
        processingPass[bkPathForDir[dirLfn]] = metadata[4]
        prodForBKPath.setdefault(bkPathForDir[dirLfn], set()).add(metadata[7])
      else:
        invisible.add(dirLfn)
    else:
      lfnsFromBK.add(longDir)

  # For those not available ask the BK
  chunkSize = 100
  if lfnsFromBK:
    gLogger.always('Getting BK metadata for %d directories' % len(lfnsFromBK))
    success = {}
    for lfns in breakListIntoChunks(lfnsFromBK, chunkSize):
      for trial in range(10, -1, -1):
        res = bkClient.getDirectoryMetadata(lfns)
        if not res['OK'] and not trial:
          gLogger.fatal("\nError getting BK metadata", res['Message'])
          DIRAC.exit(1)
        else:
          break
      success.update(res['Value'].get('Successful', {}))
    gLogger.always('Successfully obtained BK metadata for %d directories' % len(success))
    for dirLfn in dirSet - set(bkPathForDir):
      longDir = dirShort2Long[dirLong2Short[dirLfn]]
      metadata = success.get(longDir, [{}])[0]
      if metadata:
        if metadata.get('VisibilityFlag', metadata.get('Visibility', 'Y')) == 'Y':
          bkDict = BKQuery(metadata).getQueryDict()
          bkPath = makeBKPath(bkDict)
          processingPass[bkPath] = metadata['ProcessingPass']
          prodForBKPath.setdefault(bkPath, set()).add(metadata['Production'])
        else:
          invisible.add(dirLfn)
      else:
        bkPath = "Unknown-" + dirLfn
        prodForBKPath.setdefault(bkPath, set())
      if dirLfn not in invisible:
        bkPathForDir[dirLfn] = bkPath
        bkPathUsage.setdefault(bkPath, {}).setdefault('LFN', [0, 0])

  # Ignore invisible directories
  if invisible:
    gLogger.always('%d paths are ignored as invisible' % len(invisible))
    cachedInvisible.update(invisible)
    dirSet -= invisible

  # For remaining directories, get the LFN information and the PFN information SE by SE
  if dirSet:
    missingSU = set(dirLong2Short[dirLfn] for dirLfn in dirSet)
    gLogger.always('Get LFN Storage Usage for %d directories' % len(missingSU))
    for dirLfn in missingSU:
      # LFN usage
      for trial in range(10, -1, -1):
        res = suClient.getSummary(dirLfn)
        if res['OK']:
          break
        if not trial:
          gLogger.fatal('Error getting LFN storage usage %s' % dirLfn, res['Message'])
          DIRAC.exit(1)
      bkPath = bkPathForDir[dirShort2Long[dirLfn]]
      infoType = 'LFN'
      gLogger.verbose('Directory %s: %s' % (dirLfn, str(res['Value'])))
      bkPathUsage.setdefault(bkPath, {}).setdefault(infoType, [0, 0])
      bkPathUsage[bkPath][infoType][0] += sum(val.get('Files', 0) for val in res['Value'].values())
      bkPathUsage[bkPath][infoType][1] += sum(val.get('Size', 0) for val in res['Value'].values())

    # # get the PFN usage per storage type
    # Storage type is Disk, Archived, Tape and All
    gLogger.always('Check storage type and PFN usage for %d directories' % len(dirSet))
    for dirLfn in dirSet:
      for trial in range(10, -1, -1):
        res = suClient.getDirectorySummaryPerSE(dirLfn)
        if not res['OK'] and not trial:
          gLogger.fatal('Error getting storage usage per SE %s' % dirLfn, res['Message'])
          DIRAC.exit(1)
        else:
          break
      info = physicalStorageUsage.setdefault(dirLfn, {})
      for infoType in storageTypes:
        # Active type will be recorded in Disk, just a special flag
        if infoType != 'LFN' and infoType not in info:
          nf = sum(val['Files'] for se, val in res['Value'].items() if isType(se, infoType))
          size = sum(val['Size'] for se, val in res['Value'].items() if isType(se, infoType))
          info[infoType] = {'Files': nf, 'Size': size}
      for site in storageSites:
        if site not in info:
          nf = sum(val['Files']
                   for se, val in res['Value'].items()
                   if isAtSite(se, site) and isType(se, 'Disk'))
          size = sum(val['Size']
                     for se, val in res['Value'].items()
                     if isAtSite(se, site) and isType(se, 'Disk'))
          info[site] = {'Files': nf, 'Size': size}
      bkPath = bkPathForDir[dirLfn]
      for infoType in info:
        bkPathUsage.setdefault(bkPath, {}).setdefault(infoType, [0, 0])
        bkPathUsage[bkPath][infoType][0] += info[infoType].get('Files', 0)
        bkPathUsage[bkPath][infoType][1] += info[infoType].get('Size', 0)
      datasetStorage[storageType(res['Value'])].add(bkPath)

  gLogger.always('Obtained BK path and storage usage of %d directories in %.1f seconds' %
                 (len(dirSet), time.time() - startTime))


def isType(se, infoType):
  """check if an SE is of a given type."""
  if infoType == 'All':
    return True
  if infoType == 'LFN':
    return False
  return storageType([se]) == infoType


def isAtSite(se, site):
  """checks if an SE is at a given site."""
  seSite = cachedSESites.get(se)
  if seSite is None:
    seSite = dmsHelper.getLocalSiteForSE(se)
    if seSite['OK']:
      cachedSESites[se] = seSite['Value']
  return site == seSite


def prBinNumber(binNumber):
  """Return bin number as a string."""
  week = binNumber
  year = week / 52
  week = week % 52
  return 'y%4d/w%02d' % (year, week)


def prSize(size):
  """Return size as a string with sensible unit."""
  units = ('Bytes', 'kB', 'MB', 'GB', 'TB', 'PB')
  for unit in units:
    if size < 1000.:
      return '%.1f %s' % (size, unit)
    size /= 1000.
  return '???'


def getPhysicalUsage(baseDir):
  """Extract information about storage usage from the StorageusageDB."""
  for trial in range(10, -1, -1):
    res = suClient.getStorageDirectoryData(baseDir, None, None, None, timeout=3600)
    if not res['OK']:
      if not trial:
        gLogger.fatal("Error getting list of directories for %s" % baseDir, res['Message'])
        DIRAC.exit(1)
    else:
      break
  # The returned value is a dictionary of all directory leaves
  info = res['Value']
  dirSet = set()
  nf = 0
  size = 0
  gLogger.info("Storage usage OK for %d directories in %s" % (len(info), baseDir))
  for directory in info:
    nodes = directory.split('/')
    try:
      prod = int(nodes[5])
      # FIXME
      # Don't count the certification productions
      if prod < 2000 and nodes[2] == 'validation':
        continue
    except (IndexError, ValueError):
      pass
    if len(nodes) < 5 or nodes[4] not in ('LOG', 'HIST', 'BRUNELHIST', 'DAVINCIHIST', 'GAUSSHIST'):
      physicalStorageUsage.setdefault(directory, {}).setdefault('All', info[directory])
      dirSet.add(directory)
      nf += info[directory]['Files']
      size += info[directory]['Size']
  # In case the base directory was not a leave, cache its physical storage data
  physicalStorageUsage.setdefault(baseDir, {}).setdefault('All', {'Files': nf, 'Size': size})
  return dirSet


def storageType(seList):
  """Return the storage type: disk, tape or archive."""
  if not (set(se for se in seList if not dmsHelper.isSEArchive(se)) -
          set(['CERN-SWTEST', 'CERN-FREEZER-EOS', 'CERN-FREEZER'])) and \
     set(se for se in seList if dmsHelper.isSEArchive(se)):
    # Only -ARCHIVE
    return 'Archived'
  tapeSEs = resolveSEGroup('Tier1-RAW') + resolveSEGroup('Tier1-RDST')
  if [se for se in seList if se in tapeSEs]:
    # Any file on Tape
    return 'Tape'
  # Others
  return 'Disk'


def scanPopularity(since, getAllDatasets, topDirectory='/lhcb', csvFile=None):
  """That function does the job to cache the directories, get the corresponding
  datasets and join with the popularity."""
  # Reset global variables

  bkPathForDir.clear()
  cachedInvisible.clear()
  prodForBKPath.clear()
  bkPathUsage.clear()
  processingPass.clear()
  bkPathPopularity.clear()
  physicalStorageUsage.clear()
  datasetStorage.clear()
  for infoType in storageTypes:
    datasetStorage[infoType] = set()

  # set of used directories
  usedDirectories = set()
  usedSEs = {}
  binSize = 'week'
  nbBins = int((since + 6) / 7)
  since = 7 * nbBins

  from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
  ignoreDirectories = Operations().getValue('DataManagement/PopularityIgnoreDirectories',
                                            ['user', 'test', 'debug', 'dataquality', 'software',
                                             'database', 'swtest', 'certification', 'validation'])
  nowBin = getTimeBin(datetime.now() - timedelta(days=1))
  notCached = set()

  if getAllDatasets:
    # Get list of directories
    startTime = time.time()
    res = FileCatalog().listDirectory(topDirectory)
    if not res['OK']:
      gLogger.fatal("Cannot get list of directories", res['Message'])
      DIRAC.exit(1)
    directories = set(subDir for subDir in res['Value']['Successful'][topDirectory]['SubDirs']
                      if subDir.split('/')[2] not in ignoreDirectories and
                      'RAW' not in subDir and 'RDST' not in subDir and 'SDST' not in subDir)
    allDirectoriesSet = set()
    for baseDir in directories:
      allDirectoriesSet.update(getPhysicalUsage(baseDir))
    gLogger.always("Obtained %d directories storage usage information in %.1f seconds)" %
                   (len(allDirectoriesSet), time.time() - startTime))
    cacheDirectories(allDirectoriesSet)

  # Get the popularity raw information for the specified number of days
  if since:
    entries = 0
    now = datetime.now()
    endTime = datetime(now.year, now.month, now.day, 0, 0, 0)
    startTime = endTime
    gLogger.always('Get popularity day-by-day')
    stTime = time.time()
    for _i in range(since):
      endTime = startTime
      startTime = endTime - timedelta(days=1)
      endTimeQuery = endTime.isoformat()
      startTimeQuery = startTime.isoformat()
      status = 'Used'
      # Getting the popularity with 10 retries for the given day
      for _i in range(10):
        res = duClient.getDataUsageSummary(startTimeQuery, endTimeQuery, status, timeout=7200)
        if res['OK']:
          break
        gLogger.error("Error getting popularity entries, retrying...", res['Message'])
      if not res['OK']:
        gLogger.fatal("Error getting popularity entries", res['Message'])
        DIRAC.exit(1)
      val = res['Value']
      entries += len(val)

      # Get information on useful directories
      directories = set(row[1] for row in val if row[1].split('/')[2] not in ignoreDirectories)
      usedDirectories.update(directories)
      cacheDirectories(directories)

      # Get information in bins (day or week)
      for _rowId, dirLfn, se, count, insertTime in val:
        if dirLfn not in directories:
          # print rowId, dirLfn, count, insertTime, 'ignored'
          continue
        # get the binNumber (day or week)
        binNumber = getTimeBin(insertTime)
        bkPath = bkPathForDir.get(dirLfn)
        if not bkPath:
          if dirLfn not in notCached:
            notCached.add(dirLfn)
            gLogger.error('Directory %s was not cached' % dirLfn)
          bkPath = 'Unknown-' + dirLfn
        bkPathPopularity[bkPath][binNumber] = bkPathPopularity.setdefault(bkPath, {}).setdefault(binNumber, 0) + count
        usedSEs.setdefault(bkPath, set()).add(se)

    gLogger.always("\n=============================================================")
    gLogger.always("Retrieved %d entries from Popularity table in %.1f seconds" % (entries, time.time() - stTime))
    gLogger.always('Found %d datasets used since %d days' % (len(bkPathPopularity), since))
    counters = {}
    strangeBKPaths = set(bkPath for bkPath in bkPathPopularity
                         if not bkPathUsage.get(bkPath, {}).get('LFN', (0, 0))[0])
    if strangeBKPaths:
      gLogger.always('%d used datasets do not have an LFN count:' % len(strangeBKPaths))
      gLogger.always('\n'.join("%s : %s" % (bkPath, str(bkPathUsage.get(bkPath, {})))
                               for bkPath in strangeBKPaths))
    gLogger.always('\nDataset usage for %d datasets' % len(bkPathPopularity))
    for infoType in ('All', 'LFN'):
      for i in range(2):
        counters.setdefault(infoType, []).append(sum(bkPathUsage.get(bkPath, {}).get(infoType, (0, 0))[i]
                                                     for bkPath in bkPathPopularity))
    for bkPath in sorted(bkPathPopularity):
      if bkPath not in datasetStorage['Disk'] | datasetStorage['Archived'] | datasetStorage['Tape']:
        ses = usedSEs.get(bkPath)
        if ses is None:
          gLogger.error("BK path not in usedSEs", bkPath)
        else:
          datasetStorage[storageType(ses)].add(bkPath)
      nLfns, lfnSize = bkPathUsage.get(bkPath, {}).get('LFN', (0, 0))
      nPfns, pfnSize = bkPathUsage.get(bkPath, {}).get('All', (0, 0))
      gLogger.always('%s (%d LFNs, %s), (%d PFNs, %s, %.1f replicas)' %
                     (bkPath, nLfns, prSize(lfnSize), nPfns, prSize(pfnSize),
                      float(nPfns) / float(nLfns) if nLfns else 0.))
      bins = sorted(bkPathPopularity[bkPath])
      lastBin = bins[-1]
      accesses = sum(bkPathPopularity[bkPath][binNumber] for binNumber in bins)
      gLogger.always('\tUsed first in %s, %d accesses (%.1f%%), %d accesses during last %s %s' %
                     (prBinNumber(bins[0]), accesses,
                      accesses * 100. / nLfns if nLfns else 0.,
                      bkPathPopularity[bkPath][lastBin], binSize, prBinNumber(lastBin)))
    gLogger.always("\nA total of %d LFNs (%s), %d PFNs (%s) have been used" %
                   (counters['LFN'][0], prSize(counters['LFN'][1]),
                    counters['All'][0], prSize(counters['All'][1])))

  if getAllDatasets:
    # Consider only unused directories
    unusedDirectories = allDirectoriesSet - usedDirectories
    if unusedDirectories:
      gLogger.always("\n=============================================================")
      gLogger.always('%d directories have not been used' % len(unusedDirectories))
      # Remove the used datasets (from other directories)
      unusedBKPaths = set(bkPathForDir[dirLfn] for dirLfn in unusedDirectories
                          if dirLfn in bkPathForDir) - set(bkPathPopularity)
      # Remove empty datasets
      strangeBKPaths = set(bkPath for bkPath in unusedBKPaths
                           if not bkPathUsage.get(bkPath, {}).get('LFN', (0, 0))[0])
      if strangeBKPaths:
        gLogger.always('%d unused datasets do not have an LFN count:' % len(strangeBKPaths))
        gLogger.always('\n'.join("%s : %s" % (bkPath, str(bkPathUsage.get(bkPath, {})))
                                 for bkPath in strangeBKPaths))
      unusedBKPaths = set(bkPath for bkPath in unusedBKPaths
                          if bkPathUsage.get(bkPath, {}).get('LFN', (0, 0))[0])

      # In case there are datasets both on tape and disk, priviledge tape
      datasetStorage['Disk'] -= datasetStorage['Tape']
      gLogger.always("\nThe following %d BK paths were not used since %d days" %
                     (len(unusedBKPaths), since))
      for infoType in storageTypes[0:3]:
        gLogger.always("\n=========== %s datasets ===========" % infoType)
        unusedPaths = unusedBKPaths & datasetStorage[infoType]
        counters = {}
        for ty in ('All', 'LFN'):
          for i in range(2):
            counters.setdefault(ty, []).append(sum(bkPathUsage.get(bkPath, {}).get(ty, (0, 0))[i]
                                                   for bkPath in unusedPaths))
        for bkPath in sorted(unusedPaths):
          nLfns, lfnSize = bkPathUsage.get(bkPath, {}).get('LFN', (0, 0))
          nPfns, pfnSize = bkPathUsage.get(bkPath, {}).get('All', (0, 0))
          gLogger.always('\t%s (%d LFNs, %s), (%d PFNs, %s, %.1f replicas)' %
                         (bkPath, nLfns, prSize(lfnSize), nPfns,
                          prSize(pfnSize), float(nPfns) / float(nLfns)))
        gLogger.always("\nA total of %d %s LFNs (%s), %d PFNs (%s) were not used" %
                       (counters['LFN'][0], infoType, prSize(counters['LFN'][1]),
                        counters['All'][0], prSize(counters['All'][1])))
  else:
    unusedBKPaths = set()
    datasetStorage['Disk'] -= datasetStorage['Tape']

  # Now create a CSV file with all dataset information
  # Name, ProcessingPass, #files, size, SE type, each week's usage (before now)
  csvFile = 'popularity-%ddays.csv' % since if csvFile is None else csvFile
  gLogger.always("\n=============================================================")
  gLogger.always('Creating %s file with %d datasets' % (csvFile, len(bkPathPopularity) + len(unusedBKPaths)))
  with open(csvFile, 'w') as fd:
    title = "Name;Configuration;ProcessingPass;FileType;Type;Creation-%s;" % binSize + \
            "NbLFN;LFNSize;NbDisk;DiskSize;NbTape;TapeSize;NbArchived;ArchivedSize;" + \
            ';'.join(site.split('.')[1] for site in storageSites) + \
            ";Nb Replicas;Nb ArchReps;Storage;FirstUsage;LastUsage;Now"
    for binNumber in range(nbBins):
      title += ';%d' % (1 + binNumber)
    fd.write(title + '\n')
    teraByte = 1000. * 1000. * 1000. * 1000.
    for bkPath in sorted(bkPathPopularity) + sorted(unusedBKPaths):
      # Skip unknown datasets
      if bkPath.startswith('Unknown-'):
        continue
      # Not interested in histograms
      splitBKPath = bkPath.split('/')
      fileType = splitBKPath[-1]
      if 'HIST' in fileType:
        continue
      # Only RAW for partition LHCb may be of interest (and even...)
      if fileType == 'RAW' and not bkPath.startswith('/LHCb'):
        continue
      info = bkPathUsage.get(bkPath, {})
      # check if the production is still active
      prods = prodForBKPath[bkPath]
      res = transClient.getTransformations({'TransformationID': list(prods)})
      creationTime = datetime.now()
      active = []
      for prodDict in res.get('Value', []):
        creationTime = min(creationTime, prodDict['CreationDate'])
        if prodDict['Status'] in ('Active', 'Idle', 'Completed'):
          active.append(str(prodDict['TransformationID']))
      if active:
        gLogger.always("Active productions %s found in %s" % (','.join(sorted(active)), bkPath))
      if info['LFN'][0] == 0:
        continue
      for infoType in info:
        info[infoType][1] /= teraByte
      # Some BK paths contain a , to be replaces by a . for the CSV file!!
      config = '/'.join(splitBKPath[0:3])
      if ',' in bkPath:
        gLogger.always("BK path found with ',':", bkPath)
      # Name,Configuration,ProcessingPass, FileType
      row = '%s;%s;%s;%s' % (bkPath.replace('Real Data', 'RealData'),
                             config,
                             processingPass.get(bkPath, 'Unknown').replace('Real Data', 'RealData'),
                             fileType)
      # Type
      configTypes = {'/MC/Dev': 2, '/MC/Upgrade': 3}
      configType = configTypes.get(config, 0 if bkPath.startswith('/MC') else 1)
      row += ';%d' % configType
      # CreationTime
      row += ';%d' % (getTimeBin(creationTime))
      # NbLFN,LFNSize,NbDisk,DiskSize,NbTape,TapeSize, NbArchived,ArchivedSize
      for infoType in ('LFN', 'Disk', 'Tape', 'Archived'):
        row += ';%d;%f' % tuple(info[infoType])
      for site in storageSites:
        row += ';%f' % info[site][1]
      row += ';%f;%f' % (float(info['Disk'][0]) / float(info['LFN'][0]),
                         float(info['Archived'][0]) / float(info['LFN'][0]))
      if active:
        dsType = 'Active'
      else:
        dsType = 'Unknown'
        for infoType in storageTypes[0:3]:
          if bkPath in datasetStorage[infoType]:
            dsType = infoType
            break
      row += ';%s' % dsType
      bins = sorted(bkPathPopularity.get(bkPath, {}))
      if not bins:
        bins = [0]
      row += ';%d;%d;%d' % (bins[0], bins[-1], nowBin)
      usage = 0
      for binNumber in range(nbBins):
        usage += bkPathPopularity.get(bkPath, {}).get(nowBin - binNumber, 0)
        row += ';%d' % usage
      fd.write(row + '\n')
  gLogger.always('\nSuccessfully wrote CSV file %s' % csvFile)
